#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rabbit.h>
#include <dcdefs.h>
#include "co_exec.h"

#define TASK_STACK 400

#define NUM_TASKS 10
TASK tasks[NUM_TASKS];
unsigned long task_stats[NUM_TASKS];

int near a_putc(char c,void * xNotUsed)
{
	static char stdio_sema=0;
	while (SEM_LOCK(stdio_sema)!=1)
	{
		SEM_UNLOCK(stdio_sema);
		sleep(1);
	}
	if (c=='\n')
		SerialPutcA('\r');
	SerialPutcA(c);
	SEM_UNLOCK(stdio_sema);
	return 0;
}

//
// simple keyboard task just echoes keys to output
//
int near kbdtask(int param)
{
    int c=param; // gets rid of warning
    printf("Keyboard task startup\n");
    while (c!=27)
    {
        while (-1==(c=getchar()))
        {
			task_stats[param]++;
			sleep(10);
		}
        printf("Pressed '%c'\n",c);
    }
    return 0;
}

//
// Fast task, this task yields often so it is thread freindly
//
// this is always task1, so we optimize so it does nothing
// but switch. This is used to estimate task switch time.
//
int near fast_task(int param)
{
    printf("Fast task startup,param=%d\n",param);
    for (;;)
    {
	    yield;
		task_stats[1]++;     // this is always task 1.
	}
}

//
// task that waits for an event
// main task stes event once every 15 seconds
//
char event1;

int _nearcall event_task(int param)
{
	while (1)
	{
  		clr_event(&event1);
        if (wait_event(&event1,0xffffffffl))
			printf("Event Task %d wake on event.\n",param);
		task_stats[param]++;
 	}
}
//
// this task never yields, must use a timer switch to
// allow other tasks to run (preempt)
//
int near slow_task(int param)
{
    printf("Slow task startup, param=%d\n",param);
    while (!check_abort())
			task_stats[param]++;
    return 2222;
}

//
// generic task that wakes up periodically, multiple copies of this are run
//
int near task(int param)
{
	int i=0;
    printf("Task %d start.\n",param);
    while (!check_abort())
    {
		clr_event(&event1);
		task_stats[param]++;
        sleep(param*1000l);
		if (++i==10)
		{
			printf("Hello from task: %d\n",param);
			i=0;
		}
    }
    return 3333;
}

char inBuffer[32],outBuffer[1024];
#ifndef FAR_STACKS
#define STACK_SIZE 512
#else
#define STACK_SIZE 2048  // use division of 4096 for best use.
#endif
char local_stack[STACK_SIZE];

//
// count how many times a task yeilds, requires task updates the task_stats
//
void time_switches(void)
{
	unsigned long start;
	sleep(1); // sync with timer
	start=task_stats[1];
	sleep(1000l); // 1 second time
	start=task_stats[1]-start;
    printf("Task switches: %5ld/sec\n",start);
}

void main(void)
{
	unsigned char c1;
    int i,zzz=0;
	int slow_id,odd=0;
    WDT_DISABLE();
	if ((srMB0CR&0x7)==0x6)
	{
		// if running in fast RAM make sure 0ws
		outi(MB0CR,srMB0CR|=0xc0);
		outi(MB1CR,srMB1CR|=0xc0);
	}
	memset(task_stats,0,sizeof(task_stats));
	//
	// if running in flash, use serial portfor stdio
	//
    if (_inFlash())
        if( ! SerialInitA( BR_115200,               //  Port and rate
            SER_8BITS /*| SER_USE_D */, SER_IP1,     //  8 bits/Port C, IP 2
            inBuffer, sizeof inBuffer,          //  Input buffer and size
            outBuffer, sizeof outBuffer) )      //  Output buffer and size
        printf( "Error: Couldn't initialize Serial Port A\n" );
        else
        {
            _stdio = SerialA;                           //  Direct printf to SerialA
            _stdio->__out=a_putc;
        }
	// sleep requires the timer to be running
    startTimer(10,NULL,1);
    ipset0();
    printf("Startup.\n");
#ifdef FAR_STACKS
	//
	// FAR_STACKS: Separate Stack segment for each thread. Saves root space.
	// link with coexec_fs.lib
	// init_tasks() must be called before any xallocs so it get 4k alighed stacks.
	// also eact task must allocate stacks that do not cross 4k boundaries. I.e.
	// task 1=2048, task 2=1024, task 3=1024, task 5= any
	// max size is 4k per thread.
	//
    if (init_tasks(NUM_TASKS,16384l)) // number of tasks and size to allocate for all stacks
	{
		printf("init tasks failed!");
		while (SerialSendCountA());
		return;
	}
	// debug info, should show new DATASEG and SEGSIZE if far stacks used
	printf("SEGSIZE=0x%02x, STACKSEG=0x%02x, DATASEG=0x%02x\n",ini(SEGSIZE),ini(STACKSEG),ini(DATASEG));
#else
    // near stacks, just tell lin how many entries in tasks
    init_tasks(NUM_TASKS);
#endif
	i=1; // used for param, this demo sends a different param to each thread.
	// create tasks, pass near stack, size and param
    create_task(fast_task,local_stack,STACK_SIZE,i++);
	// test task switching speed
	time_switches();
	// create rest of threads,
	// use NULL to use far stacks (FAR_STACKS defined and co_exec_fs.lib)
	// or allocate from main stack (co_exec_.lib)
    create_task(task,NULL,STACK_SIZE,i++);
    create_task(task,NULL,STACK_SIZE,i++);
    create_task(task,NULL,STACK_SIZE,i++);
    slow_id=create_task(slow_task,NULL,STACK_SIZE,i++);
    create_task(event_task,NULL,STACK_SIZE,i++);
    if (_inFlash())
        create_task(kbdtask,NULL,STACK_SIZE,i++);
	// show free xmem
	printf("Free xmem: %ld\n",xavail());
    //
	// use task_switch from timer thread to allow preemption.
	// this is required in this demo because use of slow_task.
	// since slow-task never yeilds, preemption is needed.
	// However, avoid this if possible. Fully cooperative task
	// switching prevents many multiple-access problems with data
	// and may cause problems with the TCP/IP stack.
	//
    startTimer(20,task_switch,1); // force task switch every 50ms
    while (1)
    {
            sleep(1000l);
			task_stats[0]++;
            if (++zzz==20)
            {
                zzz=0;
				// demo of aborting and restarting a task.
				// abort requires the task to call check_abort()
                if (++odd&1)
                    task_abort(slow_id);
                else
				// only restart if you know the task has exited
                    restart_task(slow_id,slow_id);
            }
            if (zzz==19)
            {
			    // task stacks are filled with 0xff's. The stack check
			    // allows you to see    how much is used.
                printf("Stack checks:\n");
               for (c1=0;c1<task_count;c1++)
                   printf("\t%d bytes in task %d.\n",check_stack(c1),c1);
            }
            if (zzz==15)
				set_event(&event1);
			//
			// in this demo, task_stats[] is incremented by each task to
			// count how many times they yeield. Previous versions had this in
			// the exec, but was removed to speed up task switching and reduce mem needs.
			//
            printf("Task counts -");
            for (c1=0;c1<task_count;c1++)
                printf(" %6ld",task_stats[c1]);
            putchar('\n');
    }

}
